import numpy as np


_FLOAT_EPS = np.finfo(np.float64).eps
_EPS4 = _FLOAT_EPS * 4.0


def euler2mat(euler):
    """ Convert Euler Angles to Rotation Matrix.  See rotation.py for notes """
    euler = np.asarray(euler, dtype=np.float64)
    assert euler.shape[-1] == 3, "Invalid shaped euler {}".format(euler)

    ai, aj, ak = -euler[..., 2], -euler[..., 1], -euler[..., 0]
    si, sj, sk = np.sin(ai), np.sin(aj), np.sin(ak)
    ci, cj, ck = np.cos(ai), np.cos(aj), np.cos(ak)
    cc, cs = ci * ck, ci * sk
    sc, ss = si * ck, si * sk

    mat = np.empty(euler.shape[:-1] + (3, 3), dtype=np.float64)
    mat[..., 2, 2] = cj * ck
    mat[..., 2, 1] = sj * sc - cs
    mat[..., 2, 0] = sj * cc + ss
    mat[..., 1, 2] = cj * sk
    mat[..., 1, 1] = sj * ss + cc
    mat[..., 1, 0] = sj * cs - sc
    mat[..., 0, 2] = -sj
    mat[..., 0, 1] = cj * si
    mat[..., 0, 0] = cj * ci
    return mat


def euler2quat(euler):
    """ Convert Euler Angles to Quaternions.  See rotation.py for notes """
    euler = np.asarray(euler, dtype=np.float64)
    assert euler.shape[-1] == 3, "Invalid shape euler {}".format(euler)

    ai, aj, ak = euler[..., 2] / 2, -euler[..., 1] / 2, euler[..., 0] / 2
    si, sj, sk = np.sin(ai), np.sin(aj), np.sin(ak)
    ci, cj, ck = np.cos(ai), np.cos(aj), np.cos(ak)
    cc, cs = ci * ck, ci * sk
    sc, ss = si * ck, si * sk

    quat = np.empty(euler.shape[:-1] + (4,), dtype=np.float64)
    quat[..., 0] = cj * cc + sj * ss
    quat[..., 3] = cj * sc - sj * cs
    quat[..., 2] = -(cj * ss + sj * cc)
    quat[..., 1] = cj * cs - sj * sc
    return quat


def quat2mat(quat):
    """ Convert Quaternion to Euler Angles.  See rotation.py for notes """
    quat = np.asarray(quat, dtype=np.float64)
    assert quat.shape[-1] == 4, "Invalid shape quat {}".format(quat)

    w, x, y, z = quat[..., 0], quat[..., 1], quat[..., 2], quat[..., 3]
    Nq = np.sum(quat * quat, axis=-1)
    s = 2.0 / Nq
    X, Y, Z = x * s, y * s, z * s
    wX, wY, wZ = w * X, w * Y, w * Z
    xX, xY, xZ = x * X, x * Y, x * Z
    yY, yZ, zZ = y * Y, y * Z, z * Z

    mat = np.empty(quat.shape[:-1] + (3, 3), dtype=np.float64)
    mat[..., 0, 0] = 1.0 - (yY + zZ)
    mat[..., 0, 1] = xY - wZ
    mat[..., 0, 2] = xZ + wY
    mat[..., 1, 0] = xY + wZ
    mat[..., 1, 1] = 1.0 - (xX + zZ)
    mat[..., 1, 2] = yZ - wX
    mat[..., 2, 0] = xZ - wY
    mat[..., 2, 1] = yZ + wX
    mat[..., 2, 2] = 1.0 - (xX + yY)
    return np.where((Nq > _FLOAT_EPS)[..., np.newaxis, np.newaxis], mat, np.eye(3))


def mat2euler(mat):
    """ Convert Rotation Matrix to Euler Angles.  See rotation.py for notes """
    mat = np.asarray(mat, dtype=np.float64)
    assert mat.shape[-2:] == (3, 3), "Invalid shape matrix {}".format(mat)

    cy = np.sqrt(mat[..., 2, 2] * mat[..., 2, 2] + mat[..., 1, 2] * mat[..., 1, 2])
    condition = cy > _EPS4
    euler = np.empty(mat.shape[:-1], dtype=np.float64)
    euler[..., 2] = np.where(condition,
                             -np.arctan2(mat[..., 0, 1], mat[..., 0, 0]),
                             -np.arctan2(-mat[..., 1, 0], mat[..., 1, 1]))
    euler[..., 1] = np.where(condition,
                             -np.arctan2(-mat[..., 0, 2], cy),
                             -np.arctan2(-mat[..., 0, 2], cy))
    euler[..., 0] = np.where(condition,
                             -np.arctan2(mat[..., 1, 2], mat[..., 2, 2]),
                             0.0)
    return euler


"""
This environment simulate a turtle bot behaviour.
"""
import math
import time

import numpy as np
from hbrl.environments.navigation_environment import NavigationEnv, MapsIndex, GoalReachingNavEnv
from gym.spaces import Box
from math import pi


class TurtleEnv(NavigationEnv):

    name = "TurtleEnv"

    def __init__(self, map_tag: MapsIndex = MapsIndex.EMPTY):
        state_space = Box(low=np.array([-pi, -2, -2]), high=np.array([pi, 2, 2]))
        actions_bounds = np.array([0.2, 1])
        action_space = Box(low=-actions_bounds, high=actions_bounds)
        super().__init__(map_tag, state_space, action_space)

    def __getattr__(self, item):
        if item.startswith("_"):
            raise AttributeError(f"accessing private attribute '{item}' is prohibited")

    def reset(self):
        """
        Publish a bool message on reset topic
        :return:
        """
        self.state = super().reset()
        self.state[-1] = 0
        return self.state

    def step(self, action):
        action = np.clip(action, self.action_space.low, self.action_space.high)
        num_sub_steps = 10
        dt = 1.0 / num_sub_steps
        for _ in np.linspace(0, 1, num_sub_steps):
            # Compute new_orientation

            self.state[3] += dt * action[0] / 2  # New angular velocity
            self.state[3] = np.clip(self.state[3], self.action_space.low[0], self.action_space.high[0])
            self.state[2] += self.state[3]  # New orientation
            if self.state[2] > pi:
                self.state[2] -= 2 * pi
            elif self.state[2] < - pi:
                self.state[2] += 2 * pi

            linear_velocity = np.clip(self.state[4] + dt * action[1],
                                      self.action_space.low[1], self.action_space.high[1])  # New linear velocity
            new_state = self.state.copy()
            new_state[0] += math.cos(self.state[2]) * linear_velocity
            new_state[1] += math.sin(self.state[2]) * linear_velocity

            if self.reachable(new_state):
                self.state = new_state

            self.state[3] += dt * action[0] / 2  # New angular velocity
            self.state[3] = np.clip(self.state[3], self.action_space.low[0], self.action_space.high[0])
            self.state[2] += self.state[3]  # New orientation
            if self.state[2] > pi:
                self.state[2] -= 2 * pi
            elif self.state[2] < - pi:
                self.state[2] += 2 * pi

        return self.state, *super().get_reward_done(self.state)


class GCTurtleEnv(GoalReachingNavEnv):

    name = "GoalConditionedTurtleEnv"

    def __init__(self, map_tag: MapsIndex = MapsIndex.EMPTY, sparse_reward=True):
        turtle_env = TurtleEnv(map_tag)
        goal_space = Box(low=turtle_env.state_space.low[:2], high=turtle_env.state_space.high[:2])
        super().__init__(turtle_env, goal_space=goal_space, sparse_reward=sparse_reward)

    def get_goal_from_state(self, state):
        """
        Apply a projection P: S -> G on the given state
        """
        return state[:2].copy()

    def get_state_from_goal(self, goal):
        """
        Apply a \bar{P}:G -> S projection on the given state
        """
        return np.concatenate((goal, np.zeros(1)))



from typing import Union

import networkx
from hbrl.agents.global_planning.reachability_graph_learning.sorb import SORB
from hbrl.agents.agent import Agent
import numpy as np
import networkx as nx
from gym.spaces import Box, Discrete


class SGM(SORB):

    """
    SGM agent as defined in sparse graphical memory paper (Emmons, 2020)
    https://proceedings.neurips.cc/paper/2020/file/385822e359afa26d52b5b286226f2cea-Paper.pdf.
    """

    name = "SGM"

    def __init__(self, state_space: Union[Box, Discrete], action_space: Union[Box, Discrete], **params):
        """
        @param state_space: Environment's state space.
        @param action_space: Environment's action_space.
        @param params: Optional parameters.
        """

        super().__init__(state_space, action_space, **params)
        self.node_pruning_threshold = params.get("node_pruning_threshold", 2)
        self.max_edges_length = params.get("max_edges_length", 9)

        # Keep in memory the edge our agent is currently crossing, so we can prune it in case of failure.
        self.last_node_crossed = None
        self.target_node = None

    def build_graph(self):
        self.reachability_graph = networkx.DiGraph()
        states = self.control_policy.reinforcement_learning_agent.replay_buffer \
                    .sample(self.graph_building_buffer_size)[0][:, :2].detach().cpu().numpy()

        # states = np.array([environment.reset()[0] for _ in range(self.nb_nodes_in_graph)])
        distances = self._get_pairwise_dist(states)

        # Select nodes to add as defined in
        added_states = []
        for state_id, state in enumerate(states):
            for node_state_id, node_state in added_states:
                dist_1 = distances[state_id * states.shape[0] + node_state_id]
                dist_2 = distances[node_state_id * states.shape[0] + state_id]
                if dist_1 < self.node_pruning_threshold or dist_2 < self.node_pruning_threshold:
                    break
            else:

                self.reachability_graph.add_node(state_id, state=state.copy())

                # add edges from this new_state
                for node_state_id, node_state in added_states:
                    dist = distances[state_id * states.shape[0] + node_state_id]
                    if dist < self.max_edges_length:
                        self.reachability_graph.add_edge(state_id, node_state_id)

                    dist = distances[node_state_id * states.shape[0] + state_id]
                    if dist < self.max_edges_length:
                        self.reachability_graph.add_edge(node_state_id, state_id)

                # Add our new state to the list of added states
                added_states.append((state_id, state))

        self.graph_up_to_date = True

    def start_episode(self, state: np.ndarray, goal: np.ndarray, test_episode=False):
        Agent.start_episode(self, state, test_episode=test_episode)

        # Reset attributes that needs to
        self.current_goal = goal
        self.done = False
        self.nb_sub_task_interaction = 0
        self.last_node_crossed = None
        self.target_node = None

        if self.graph_up_to_date:
            current_node = self.get_node_for_state(state)
            final_node = self.get_node_for_state(self.current_goal, reachable_from=current_node)
            self.sub_goals = self.shortest_path(current_node, final_node)
            if self.sub_goals:
                self.next_goal = nx.get_node_attributes(self.reachability_graph, "state")[self.sub_goals[0]]
                self.target_node = self.sub_goals[0]
                assert isinstance(self.target_node, int)
            else:
                self.next_goal = self.current_goal
        else:
            self.next_goal = self.current_goal
        self.control_policy.start_episode(state, self.next_goal, test_episode=test_episode or self.graph_up_to_date)

    def process_interaction(self, action, reward, new_state, done, learn=True):
        Agent.process_interaction(self, action, reward, new_state, done, learn=not self.graph_up_to_date)
        pruned_edge = None
        if self.graph_up_to_date:
            if self.reached_sub_goal(new_state, self.next_goal):
                if self.target_node is not None:
                    self.last_node_crossed = self.target_node
                if self.sub_goals:
                    self.sub_goals.pop(0)
                if self.sub_goals:
                    self.next_goal = nx.get_node_attributes(self.reachability_graph, "state")[self.sub_goals[0]]
                    self.target_node = self.sub_goals[0]
                    assert isinstance(self.target_node, int)
                else:
                    self.next_goal = self.current_goal
                    self.target_node = None  # Keep in memory that we are not crossing an edge. Prevent pruning failures.
                self.nb_sub_task_interaction = 0
                self.control_policy.start_episode(new_state, self.next_goal,
                                                  test_episode=self.under_test or self.graph_up_to_date)
            else:
                self.nb_sub_task_interaction += 1
                if self.nb_sub_task_interaction >= self.max_interactions_per_sub_task:

                    """ This test and everything inside is the pruning procedure, specific to SGM. """
                    if learn and self.last_node_crossed is not None and self.target_node is not None:

                        # Prune the edge we tried to cross
                        self.reachability_graph.remove_edge(self.last_node_crossed, self.target_node)
                        pruned_edge = (self.last_node_crossed, self.target_node)
                        # Re-planing
                        self.start_episode(new_state, self.current_goal, test_episode=self.under_test)

                        # RE-PLANNING LOOP STOP CASE:
                        #   We can be sure that each episode finally stop because it exists at least one path in the
                        # graph our agent can cross entirely: the empty path.
                        # If every edge are not reachable, we will remove all of them and our agent will end up by
                        # trying to reach the final goal directly, which will lead to a re-planing in case of failure
                        # because self.last_node_crossed and self.target_node will be None.
                    else:
                        self.done = True
                assert (self.next_goal == self.control_policy.current_goal).all()
        else:
            self.control_policy.process_interaction(action, reward, new_state, done, learn=learn)
        assert (self.next_goal == self.control_policy.current_goal).all()
        return pruned_edge




import copy
import math
from random import random
from typing import Union

import networkx as nx
import numpy as np
from gym.spaces import Discrete, Box
from networkx import NetworkXNoPath, NetworkXError
from hbrl.agents.agent import Agent
from hbrl.agents.value_based_agent import ValueBasedAgent
from hbrl.agents.goal_conditioned_wrappers.tilo import TILO


class RGL(Agent):
    """
    RGL stands for Reachability Graph Learning
    An agent that is trained to learn the environment reachability_graph, so that learn by interacting with its
    environment, but don't need to reach a goal to do so. Then, he is able to exploit his knowledge of his environment
    to reach goals or to patrol inside it.
    self.mode is used to make the agent know what he should do during this episode.
    """

    name = "RGL"

    def __init__(self, goal_conditioned_wrapper, value_based_agent_class,
                 state_space: Union[Box, Discrete], action_space: Union[Box, Discrete], **params):
        """
        @param goal_conditioned_wrapper: Wrapper used to turn the given value based agent class into a goal
            conditioned agent.
        @param value_based_agent_class: Value based agent class.
        @param state_space: Environment's state space.
        @param action_space: Environment's action_space.
        @param params: Optional parameters.
        """

        super().__init__(state_space, action_space, **params)

        assert issubclass(goal_conditioned_wrapper, TILO)
        assert issubclass(value_based_agent_class, ValueBasedAgent)
        self.control_policy_goal_conditioned_wrapper = goal_conditioned_wrapper
        self.control_policy_value_based_agent_class = value_based_agent_class

        # ALGORITHM TYPE
        # re_usable_policy=True implies a pre-training to the given policy, and that this policy will not be trained
        # during learning..
        self.re_usable_policy = params.get("re_usable_policy", True)

        # Margin distance under which a goal can be considered reached
        # Ex.: s = (x, y), g = (x', y'), tolerance_margin = (a, b)
        # goal is reached by state s if x' - x <= a and y' - y <= b.
        self.tolerance_radius = params.get("tolerance_radius", .8)

        # GRAPH BUILDING HYPER-PARAMETERS
        self.nodes_attributes = params.get("nodes_attributes", {})
        self.edges_attributes = params.get("edges_attributes", {})
        #  '--> Default nodes and edges attributes on creation as a dict, like {nb_explorations: 0} for nodes.

        self.exploration_duration = params.get("exploration_duration", 90)
        #  '--> Duration of any explorations from a node we want to explore from, in number of interactions.

        # SUB-GOALS PLANNING ARGUMENTS
        self.max_steps_to_reach = params.get("max_steps_to_reach", 50)

        # MISC. ARGUMENTS
        self.verbose = params.get("verbose", False)
        # Copy params and store them in self
        self.init_params = {}
        for k, v in params.items():
            self.init_params[k] = copy.deepcopy(v) if k != "goal_reaching_agent" else v.copy()

        # LOW-LEVEL ACTIONS PLANNING ARGUMENTS

        """
        General attributes (for any mode)
        """
        self.directed_graph = params.get("directed_graph", True)
        self.reachability_graph = nx.DiGraph() if self.directed_graph else nx.Graph()
        self.learn = True
        self.default_state = params.get("default_state")
        assert self.default_state is not None

        # Compute out goal space
        self.goal_space = params.get("goal_space", self.state_space)
        assert isinstance(self.goal_space, Box) or isinstance(self.goal_space, Discrete)
        self.goal_size = self.goal_space.shape[0]
        self.goal_shape = self.goal_space.shape
        assert len(self.goal_shape) == 1, "Multi dimensional spaces are not supported."

        # Compute state_to_goal_filter, (state[self.state_to_goal_filter] = <goal associated to state>) aka a projection
        default_filter = np.array([1] * self.goal_size + [0] * (self.state_size - self.goal_size)).astype(bool)
        self.state_to_goal_filter = params.get("state_to_goal_filter", default_filter)
        assert np.argwhere(self.state_to_goal_filter).shape[0] == self.goal_size

        self.control_policy: TILO = goal_conditioned_wrapper(value_based_agent_class, state_space=self.state_space,
                                                             action_space=self.action_space, **params)

        # The number of episode will depend on the task we are doing. In any mode, the agent choose when he's done doing
        # the task.
        #   For exploration, our agent will continue its environment while he didn't finish his exploration strategy.
        #   For goal reaching mode, our agent will continue until he reached the goal or until he considers that he
        # failed to reach it.
        #   For patrol, our agent will fix a custom time limit.
        self.done = False

        # In any mode, we will try to follow trajectories inside the learned reachability graph. Doing so, if we
        # failed reaching the next node, the entire path is compromised. To make sure we don't try infinitely to reach
        # an unreachable sub-goal, we should count how many time steps we tried to reach it so far, to detect it as an
        # impossible subtask. This allows us to break and start a new episode instead of looping infinitely.
        self.current_subtask_steps = 0  # Steps done so far to reach the next node.
        self.last_node_passed = None  # So we can know which edge is concerned if it failed.
        self.next_way_point = None

        """
        EXPLORATION AND GRAPH BUILDING ATTRIBUTES
            Explorations steps are simple:
             - Select a node to explore from (the one less chosen for exploration)
             - Find a path to this node in the topological graph using A*, Dijkstra, or any path finding algorithm,
             - For each waypoint, try to reach it, until we don't.
             - Once we reached the last sub-goal (aka. The node initially selected for exploration), perform randoms
             actions for a fixed duration, to sample random states next to this node.  
             - Use the randomly sampled states to add new nodes to our graph.
        """
        # Once we reached a node that has been selected as interesting for exploration, we will explore using a random
        # policy for a fixed duration.
        self.global_path = None
        self.higher_node_id = -1  # Identifier of the younger node. To know which id to give to a new node.
        self.select_goal_for_exploration = params.get("select_goal_for_exploration", True)
        self.exploration_goal_range = params.get("exploration_goal_range", 2)
        # Trajectory made once we reached last exploration waypoint.
        self.last_exploration_trajectory = []
        self.distance_estimation_max = None
        self.edges_distance_threshold = params.get("edges_distance_threshold", 0.7)
        self.nodes_distance_threshold = params.get("nodes_distance_threshold", 0.5)
        self.sampled_exploration_goal = None
        self.under_exploration = False
        # We don't want to perform (state, goal) distance on states and goals that are too distant from each others,
        # otherwise, we can feed into the policy neural networks g - s data that are too different from the ones
        # seen in the pre-training, and get fake d_pi(g - P(s)) results. This is the maximum acceptable distance for it.
        self.s_g_max_dist = params.get("s_g_max_dist", 19)

        """
        GOAL REACHING ATTRIBUTES
        """
        self.final_goal = None

        self.exploration_node = None
        self.node_from = None

    """
    AGENT LIFECYCLE FUNCTIONS
    """
    def on_pre_training_done(self, start_state, reached_goals):
        """
        Compute the longer distance estimation over every goal that has been reached during the pre-training.
        It allows to choose reachability parameters more easily.
        """
        self.distance_estimation_max = None
        start_state_batch = np.tile(start_state, (len(reached_goals), 1))
        goals_batch = np.array(reached_goals)
        distance_estimations = self.control_policy.get_estimated_distances(start_state_batch, goals_batch)
        self.distance_estimation_max = distance_estimations.max()

    def start_episode(self, state: np.ndarray, goal=None, test_episode=False):
        """
        Initialise an episode using the given 'state' as the initial state, and the given goal as a goal to reach.
        If the goal argument is set, the agent will consider that he is under test, and will not learn anything.
            (fail to cross an edge will not lead to modify its weight). Otherwise, the agent will consider that he's
            training and will select a node to explore.
        """
        if test_episode:
            assert isinstance(goal, np.ndarray) and goal.shape == self.goal_shape
        super().start_episode(state, test_episode=test_episode)

        # Reset episodes variables if we already made an episode before
        self.last_node_passed = None
        self.global_path = None
        self.last_exploration_trajectory = []
        self.current_subtask_steps = 0
        self.done = False
        self.under_exploration = False

        # Start episode
        self.learn = goal is None
        if len(self.reachability_graph.nodes()) == 0:
            self.create_node(state)  # Create node on state with id=0 for reachability graph initialisation
        if not self.learn:
            assert isinstance(goal, np.ndarray)
            self.final_goal = goal
        self.set_global_path(state)
        p = self.global_path
        self.set_next_way_point()
        self.control_policy.start_episode(state, self.next_way_point, test_episode=True)
        if self.verbose:
            print("New episode. Learn = ", self.learn, ". Selected next node = " + str(self.next_way_point))

    def action(self, state, explore=True):
        if self.under_exploration:
            return self.action_space.sample()
        return self.control_policy.action(state, explore=False)  # Assume the control policy has been trained before.

    def process_interaction(self, action, reward, new_state, done, learn=True):
        self.under_exploration = self.learn and (
            (len(self.global_path) == 1 and self.select_goal_for_exploration) or
            (len(self.global_path) == 0 and not self.select_goal_for_exploration)
        )

        if self.under_exploration:
            self.last_exploration_trajectory.append(new_state)

        reached = False
        if self.global_path:
            reached = self.reached(new_state, self.next_way_point)
            reward = 1.0 if reached else 0.0  # Will not be used if self.re_usable_policy == True
            self.control_policy.process_interaction(action, reward, new_state, done=reached,
                                               learn=not self.re_usable_policy)

        if self.global_path and reached:
            # The next sub-goal have been reached, we can remove it and continue to the next one
            self.current_subtask_steps = 0
            self.control_policy.stop_episode()
            passed_node = self.global_path.pop(0)
            if isinstance(passed_node, int):
                if self.last_node_passed is not None and self.learn:
                    self.on_edge_crossed(self.last_node_passed, passed_node)
                self.last_node_passed = passed_node
            if self.global_path:
                # Get next goal
                self.set_next_way_point()
                self.control_policy.start_episode(new_state, self.next_way_point, test_episode=True)
            elif learn and self.select_goal_for_exploration:
                self.extend_graph()
                self.done = True
                # else, agent continue while the main don't consider we reached the goal.

            if self.verbose:
                if len(self.global_path) == 1:
                    print("Path is done," + " exploring." if self.learn else " reaching final goal.")
                else:
                    print("Reached a way point. Next one is " + str(self.global_path[0]) + ".")
        else:
            self.current_subtask_steps += 1
            max_subtask_steps = self.exploration_duration if self.under_exploration else self.max_steps_to_reach

            if self.current_subtask_steps >= max_subtask_steps:
                if self.global_path and isinstance(self.global_path[0], int):
                    self.on_edge_failed(self.last_node_passed, self.global_path[0])

                if self.under_exploration:
                    self.extend_graph()
                    if self.select_goal_for_exploration:
                        self.control_policy.stop_episode()
                self.done = True
                if self.verbose:
                    print("We failed reaching this way point ... We're done with this episode.")
            elif self.verbose:
                print("Trying to reach way point " + str(self.next_way_point) + ". Time steps left = "
                      + str(max_subtask_steps - self.current_subtask_steps))

        # Increment the counter of the node related to 'new_state'.
        super().process_interaction(action, reward, new_state, done)
        if self.verbose:
            print("Interaction: observation=" + str(self.last_state) + ", action=" + str(action) + ", new_state="
                  + str(new_state) + ", agent goal=" + str(self.control_policy.current_goal))

    def set_next_way_point(self):
        if not self.global_path:
            self.next_way_point = None
        elif isinstance(self.global_path[0], np.ndarray):
            self.next_way_point = self.global_path[0].copy()
        elif isinstance(self.global_path[0], int):
            self.next_way_point = self.get_node_attribute(self.global_path[0], "state").copy()

    def on_edge_crossed(self, last_node_passed, next_node_way_point):
        if not isinstance(last_node_passed, int) or not isinstance(next_node_way_point, int):
            return

    def on_edge_failed(self, last_node_passed, next_node_way_point):
        if not isinstance(last_node_passed, int) or not isinstance(next_node_way_point, int):
            return
        edge_attributes = self.reachability_graph.edges[last_node_passed, next_node_way_point]
        edge_attributes["cost"] = float("inf")

    """
    PLANNING FUNCTIONS 
    """
    def reached(self, state: np.ndarray, goal: np.ndarray) -> bool:
        return np.linalg.norm(state[self.state_to_goal_filter] - goal) < self.tolerance_radius

    def shortest_path(self, node_from, node_to_reach):
        try:
            return nx.shortest_path(self.reachability_graph, node_from, node_to_reach, "cost")
        except NetworkXNoPath:
            return []

    def get_path_to(self, state, goal) -> list:
        """
        Use the information stored about the environment to compute a global path from the given state to the given
        goal.
        """
        node_from = self.get_node_for_state(state)
        node_to = self.get_node_for_state(goal)
        return self.shortest_path(node_from, node_to)

    def sample_exploration_target(self, explored_node):

        angle = random() * 2 * math.pi - math.pi
        if isinstance(explored_node, int):
            last_node_explored_state = self.get_node_attribute(explored_node, "state")
        else:
            last_node_explored_state = explored_node
        # Select a target
        x_diff = math.cos(angle) * self.exploration_goal_range
        y_diff = math.sin(angle) * self.exploration_goal_range
        target_goal = last_node_explored_state.copy()
        target_goal[:2] += [x_diff, y_diff]
        self.sampled_exploration_goal = target_goal.copy()
        return target_goal

    """
    ENVIRONMENT EXPLORATION / GRAPH BUILDING FUNCTIONS
    """
    def set_global_path(self, state):
        """
        Set the global path. It will depend on whether we are exploring or not.
        """
        self.global_path = []
        if not self.reachability_graph.nodes:
            raise Exception("Agent's graph hasn't been initialised properly.")

        # Get the node with the lowest number of explorations from it, or reached counter.
        global_path = None
        self.node_from = self.get_node_for_state(state)
        if self.learn:
            # Find the closest node g_0 as our first sub_goal
            g_0_state = nx.get_node_attributes(self.reachability_graph, "state")[self.node_from]
            distance_to_g0 = self.get_normalised_distance_estimation(state, g_0_state)
            if distance_to_g0 > self.edges_distance_threshold:
                self.global_path = []  # Impossible to join another graph than our current one.
            else:
                self.exploration_node = None
                # Sort nodes by how many times we explored from it.
                nodes_nb_explorations = nx.get_node_attributes(self.reachability_graph, "explorations")
                sorted_by_explorations = sorted(nodes_nb_explorations.items(), key=lambda item: item[1])
                nodes_sorted_by_explorations = [elt[0] for elt in sorted_by_explorations]
                # For each node, from the less explored to the most, verify if it is reachable without unreachable edge
                for node_id in nodes_sorted_by_explorations:
                    global_path = self.shortest_path(self.node_from, node_id)
                    if not global_path:
                        # No path to this node, let's check the next one.
                        continue

                    # We found a path to the given node. Iterate through it to verify if it contains an unreachable edge.
                    node_1 = global_path[0]
                    for node_2 in global_path[1:]:
                        if nx.get_edge_attributes(self.reachability_graph, "cost")[(node_1, node_2)] == float("inf"):
                            # This edge is unreachable. Stop iterating through this path.
                            # NB: break here will prevent to enter the else statement bellow.
                            break
                        node_1 = node_2
                    else:
                        # The path contains no unreachable edges. we can select this node for exploration and try to
                        # reach it using the path we found.
                        self.exploration_node = node_id
                        self.reachability_graph.nodes[self.exploration_node]["explorations"] += 1
                        break
                if self.exploration_node is None:
                    self.exploration_node = min(self.reachability_graph.nodes(data=True), key=lambda x: x[1]["explorations"])[0]
                    self.reachability_graph.nodes[self.exploration_node]["explorations"] += 1

                # Find the shortest path through our reachability graph to the best node
                self.global_path = self.shortest_path(self.node_from, self.exploration_node) if global_path is None else global_path
        else:
            node_from = self.get_node_for_state(state)
            target_node = self.get_node_for_state(self.final_goal)

            # Find the shortest path through our reachability graph to the best node
            self.global_path = self.shortest_path(node_from, target_node) if global_path is None else global_path

        if self.learn and self.select_goal_for_exploration:
            self.global_path.append(self.sample_exploration_target(self.global_path[-1] if self.global_path else self.last_state))
        elif not self.learn:
            self.global_path.append(self.final_goal)

    """
    GRAPH EXPLOITATION FUNCTIONS
    """
    def get_node_attribute(self, node_id, attribute):
        attributes = nx.get_node_attributes(self.reachability_graph, attribute)
        return attributes.get(node_id)

    def get_node_for_state(self, state, data=False):
        """
        Select the node that best represent the given state.
        """
        assert isinstance(state, np.ndarray)
        if state.shape[-1] == len(self.state_to_goal_filter):
            state = state[self.state_to_goal_filter]
        if not self.reachability_graph.nodes:
            return None  # The graph  hasn't been initialised yet.
        node_data = None
        closest_distance = None
        for node_id, args in self.reachability_graph.nodes(data=True):
            distance = self.control_policy.get_estimated_distances(state, args["state"])
            if closest_distance is None or distance < closest_distance:
                node_data = (node_id, args)
                closest_distance = distance
        return node_data if data else node_data[0]

    def get_normalised_distance_estimation(self, states, goals):
        """
        Use the UVFA to get a value function approximation between two states.
        """

        # Convert inputs to batch
        if len(states.shape) == 1:
            states = states[np.newaxis]
        if len(goals.shape) == 1:
            goals = goals[np.newaxis]

        # Operate projections: States should be in the state space and goals in the sub-goal space.
        # - Inverse projection G -> S if the states don't have the right size
        #   (computing distance between goal and state).
        if states.shape[-1] != self.state_size:
            if states.shape[-1] != self.goal_size:
                raise ValueError("Unknown state shape.")
            states = self.goal_to_state(states)

        # - Projection S -> G if the given sub-goals are actually states or goals. Our control policy only know
        #   goals, not goals.
        if goals.shape[-1] != self.goal_size:
            if goals.shape != self.state_size:
                raise ValueError("Unknown goal shape")
            goals = goals[self.state_to_goal_filter]

        # Now inputs have the right size. We can compute the distance.
        estimated_distance = self.control_policy.get_estimated_distances(states, goals)

        # Returns the normalised estimated distance (assuming self.distance_estimation_min == 0)
        assert isinstance(estimated_distance / self.distance_estimation_max, np.ndarray)
        return estimated_distance / self.distance_estimation_max

    def get_reachable_from(self, node):
        """
        return a set of nodes that are reachable from the given node
        :param node:
        :return:
        """
        return list(self.reachability_graph.neighbors(node))

    """
    GRAPH BUILDING FUNCTIONS
    Some of them can/should be override by subclasses that implement different graph building strategies.
    """
    def create_node(self, state, **params):
        assert len(state.shape) == 1, "A node cannot be created from states batch"
        attributes = copy.deepcopy(self.nodes_attributes)
        attributes["explorations"] = 0
        for key, value in params.items():
            attributes[key] = value
        for key, value in attributes.items():
            if isinstance(value, tuple) and len(value) == 2 and callable(value[0]):
                # Here, the value of this parameter should be initialised using a function call.
                # The value inside self.nodes_attributes is a tuple, with the function in first attribute, and it's
                # parameters as a dict in the second.
                # Ex: self.create_node(weights, {model: (initialise_neural_network, {layer_1_size: 200})}
                #   will do attributes[model] = initialise_neural_network(layer_1_size=200)
                function = value[0]
                parameters_dict = value[1]
                attributes[key] = function(**parameters_dict)

        node_id = self.higher_node_id + 1
        self.higher_node_id += 1
        attributes["state"] = state[self.state_to_goal_filter]
        # NB: State is the state that belong to this node. Because it's a topological graph, every node have a position
        # in the state space that can be associated with a state. In neural gas, the word 'weights' is used in reference
        # to neurons weights. But we prefer to use the word 'state' to avoid any ambiguity with path finding cost of
        # taking a node in path finding algorithms. We use the word 'cost' on edges for that.
        self.reachability_graph.add_node(node_id, **attributes)
        return node_id

    def create_edge(self, first_node, second_node, **params):
        """
        Create an edge between the two given nodes. If potential is True, it means that the edge weight will be lower in
        exploration path finding (to encourage the agent to test it) or higher in go to mode (to bring a lower
        probability of failure).
        """
        assert (first_node, second_node) not in self.reachability_graph.edges
        attributes = copy.deepcopy(self.edges_attributes)
        for key, value in params.items():
            attributes[key] = value
        attributes["cost"] = params.get("cost", 1.)  # Set to one only if it's unset.
        self.reachability_graph.add_edge(first_node, second_node, **attributes)

    def extend_graph(self):
        """
        Update the reachability graph using the exploration trajectory.
        Precondition: An exploration trajectory has been made.
        """
        assert not self.under_test
        assert self.last_exploration_trajectory != []

        # Build nodes + explored states batch
        explored_states = np.array(self.last_exploration_trajectory)
        explored_goals = explored_states[:, self.state_to_goal_filter]
        nodes_goals = np.array(list(nx.get_node_attributes(self.reachability_graph, "state").values()))
        nodes_states = self.goal_to_state(nodes_goals)
        nb_explored_states = len(self.last_exploration_trajectory)
        nb_nodes = nodes_states.shape[0]

        explored_states = np.repeat(explored_states, nb_nodes, axis=0)
        explored_goals = np.tile(explored_goals, (nb_nodes, 1))
        nodes_states = np.repeat(nodes_states, nb_explored_states, axis=0)
        nodes_goals = np.tile(nodes_goals, (nb_explored_states, 1))

        forward_distance_estimations = self.get_normalised_distance_estimation(nodes_states, explored_goals)
        backward_distance_estimations = self.get_normalised_distance_estimation(explored_states, nodes_goals)
        for state_id, state in enumerate(self.last_exploration_trajectory):
            links_to_do = []

            # Select relevant nodes for this state (that are not too far so the computed distance is relevant, because
            # data are not too different from what our agent saw in his pre-training).

            # Get close node's ID
            nodes_goals_ = np.array(list(nx.get_node_attributes(self.reachability_graph, "state").values()))
            relevant_nodes_ids = \
                np.squeeze(np.argwhere((np.absolute(nodes_goals_ - state) <= self.s_g_max_dist).all(-1)), -1).tolist()

            # Iterate through them
            # for node_id in self.reachability_graph.nodes(data=False):
            for node_id in relevant_nodes_ids:

                forward_estimated_distance = forward_distance_estimations[node_id * nb_explored_states + state_id]
                backward_estimated_distance = backward_distance_estimations[state_id * nb_nodes + node_id]

                if forward_estimated_distance < self.nodes_distance_threshold or \
                        backward_estimated_distance < self.nodes_distance_threshold:
                    break
                if forward_estimated_distance < self.edges_distance_threshold:
                    links_to_do.append({"node": node_id, "forward": True, "cost": forward_estimated_distance})
                if backward_estimated_distance < self.edges_distance_threshold:
                    links_to_do.append({"node": node_id, "forward": False, "cost": backward_estimated_distance})
            else:
                # => this observation is far enough from any nodes
                # Create node
                new_node = self.create_node(state)

                new_node_state_batch = np.repeat(state[np.newaxis], nb_explored_states, axis=0)
                new_node_goal_batch = new_node_state_batch[:, self.state_to_goal_filter]
                explored_states_array = np.array(self.last_exploration_trajectory)

                new_node_forward_distance_estimations = self.get_normalised_distance_estimation(
                    new_node_state_batch, explored_states_array[:, self.state_to_goal_filter])
                new_node_backward_distance_estimations = \
                    self.get_normalised_distance_estimation(explored_states_array, new_node_goal_batch)

                forward_distance_estimations = np.concatenate((forward_distance_estimations,
                                                               new_node_forward_distance_estimations))
                backward_distance_estimations = np.insert(backward_distance_estimations,
                                                          np.arange(1, nb_explored_states + 1) * nb_nodes,
                                                          new_node_backward_distance_estimations)
                nb_nodes += 1

                for link_to_do in links_to_do:
                    if link_to_do["forward"]:
                        self.create_edge(new_node, link_to_do["node"], cost=link_to_do["cost"])
                    else:
                        self.create_edge(link_to_do["node"], new_node, cost=link_to_do["cost"])

    def remove_graph_around(self, node):
        """
        Remove the sub-graph around the given node. Should be called if a part of the graph is isolated.
        """
        to_remove = [n for n in self.reachability_graph.neighbors(node)]
        self.reachability_graph.remove_node(node)
        for node in to_remove:
            self.remove_graph_around(node)

    def remove_node(self, node_id, remove_isolated_nodes=True):
        """
        Remove a node in the graph. Ensure that every node are still reachable from the initial node.
        :return: list of nodes additionally removed
        """
        neighbors = self.reachability_graph.neighbors(node_id)
        self.reachability_graph.remove_node(node_id)
        if remove_isolated_nodes:
            removed_nodes = []
            for node_id in copy.deepcopy(neighbors):
                try:
                    nx.shortest_path(self.reachability_graph, 0, node_id)
                except NetworkXNoPath:
                    removed_nodes.append(node_id)
                    self.remove_graph_around(node_id)
            return removed_nodes
        return []

    def remove_edge(self, node_1, node_2):
        # Remove the edge
        try:
            self.reachability_graph.remove_edge(node_1, node_2)
        except NetworkXError:
            return

        # If this operation isolated a part of the graph from the start point, remove the isolated sub-graph
        try:
            nx.shortest_path(self.reachability_graph, 0, node_1)
        except NetworkXNoPath:
            self.remove_graph_around(node_1)
        try:
            nx.shortest_path(self.reachability_graph, 0, node_2)
        except NetworkXNoPath:
            self.remove_graph_around(node_2)

    def goal_to_state(self, goals):
        goals_ = goals[np.newaxis] if len(goals.shape) == 1 else goals
        states = np.tile(self.default_state, (goals_.shape[0], 1))
        states[:, self.state_to_goal_filter] = goals_
        return goals_

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == "goal_reaching_agent":
                setattr(result, k, self.control_policy.copy())
            elif k == "init_params":
                new_dict = {}
                for k_, v_ in v.items():
                    new_dict[k_] = copy.deepcopy(v_) if k_ != "goal_reaching_agent" else v_.copy()
                setattr(result, k, new_dict)
            else:
                setattr(result, k, copy.deepcopy(v, memo))
        return result

    def reset(self):
        self.__init__(self.control_policy_goal_conditioned_wrapper, self.control_policy_value_based_agent_class,
                      self.state_space, self.action_space, **self.init_params)


from enum import Enum

class AgentsIndex(Enum):
    RGL = "agents.graph_planning.tc_rgl.TC_RGL"
    REO_RGL = "agents.graph_planning.reo_rgl.REO_RGL"
    TC_RGL = "agents.graph_planning.rgl.RGL"
    DQN = "agents.discrete.goal_conditioned_dqn_her_diff.DqnHerDiffAgent"
    SAC = "agents.continuous.sac.SAC"
    SORB = "agents.global_planning.reachability_graph_learning.sorb"
    SGM = "agents.global_planning.reachability_graph_learning.sgm"



import numpy as np
from gym.spaces import Box
from hbrl.agents.goal_conditioned_wrappers.her import HER


class TILO(HER):
    """
    A global agent class for goal conditioned agents. The # NEW tag indicate differences between Agent class and this
    one.
    """

    def __init__(self, reinforcement_learning_agent_class, state_space, action_space, **params):
        assert isinstance(state_space, Box), "The state space should be an instance of gym.spaces.Box. " \
                                             "Discrete state space is not supported."

        super().__init__(reinforcement_learning_agent_class, state_space, action_space, **params)
        self.name = self.reinforcement_learning_agent.name + " + TILO"

    @property
    def feature_space(self):
        if isinstance(self.state_space, Box):
            return Box(low=self.state_space.low - self.state_space.high,
                       high=self.state_space.high - self.state_space.low)
        else:
            return self.state_space

    def get_features(self, states, goals):

        if len(states.shape) == 1 and len(goals.shape) == 2:
            states = np.repeat(states, goals.shape[0]).reshape(states.shape[0], goals.shape[0])
        elif len(goals.shape) == 1 and len(states.shape) == 2:
            goals = np.repeat(goals, states.shape[0]).reshape(goals.shape[0], states.shape[0])
        assert states.shape[:-1] == goals.shape[:-1]

        features = states.copy()
        if len(states.shape) == 1:
            state_goal_diff = goals - states[self.state_to_goal_filter]
            features[self.state_to_goal_filter] = state_goal_diff
        else:
            state_goal_diff = goals - states[:, self.state_to_goal_filter]
            features[:, self.state_to_goal_filter] = state_goal_diff
        return features



import abc
from abc import ABC
from typing import Union

import numpy as np
from gym.spaces import Box, Discrete
from hbrl.environments.environment import Environment


class GoalReachingEnv(Environment, ABC):

    name = ""

    def __init__(self, wrapped_environment, goal_space: Union[None, Box, Discrete]=None,
                 sparse_reward=True, reachability_threshold=1):
        assert isinstance(wrapped_environment, Environment)
        self.wrapped_environment = wrapped_environment
        self.goal_space = self.state_space if goal_space is None else goal_space
        assert isinstance(self.goal_space, (Box, Discrete))
        if isinstance(self.goal_space, Box):
            if isinstance(reachability_threshold, list):
                reachability_threshold = np.array(reachability_threshold)
            if isinstance(reachability_threshold, np.ndarray):
                assert self.reachability_threshold.shape[0] == self.goal_size
            self.reachability_threshold = reachability_threshold
        self.sparse_reward = sparse_reward
        self.goal = None
        self.reset()

    def __getattr__(self, name):
        """Returns an attribute with ``name``, unless ``name`` starts with an underscore."""
        if name.startswith("_"):
            raise AttributeError(f"accessing private attribute '{name}' is prohibited")
        return getattr(self.wrapped_environment, name)

    @property
    def goal_size(self):
        return self.goal_space.shape[0]

    def sample_goal(self):
        return self.goal_space.sample()

    def reset(self):
        self.wrapped_environment.reset()
        self.goal = self.sample_goal()
        return self.state, self.goal

    def step(self, action):
        state, reward, done = self.wrapped_environment.step(action)
        if done:
            return state, reward, done

        if self.reached(state):
            return state, 0, True
        else:
            reward = self.reward(state, action)
            return state, reward, False

    def reached(self, state, goal=None) -> bool:
        if goal is None:
            goal = self.goal
        state_goal = self.get_goal_from_state(state)
        if isinstance(self.goal_space, Discrete):
            return state_goal == goal
        elif isinstance(self.reachability_threshold, np.ndarray):
            return (np.abs(state_goal - goal) <= self.reachability_threshold).all()
        else:
            return np.linalg.norm(state_goal - goal) <= self.reachability_threshold

    def reward(self, state, action, goal=None):
        goal = self.goal if goal is None else goal

        if self.sparse_reward:
            return 0 if self.reached(state, goal) else -1
        else:
            state_goal = self.get_goal_from_state(state)
            return - np.linalg.norm((state_goal - goal))

    def get_goal_from_state(self, state):
        """
        Apply a projection P: S -> G on the given state
        """
        return state

    def get_state_from_goal(self, goal):
        """
        Apply a \bar{P}:G -> S projection on the given state
        """
        return goal


# Goal conditioned deep Q-network
from ..utils import scale_tensor
from ..value_based_agent import ValueBasedAgent
from ..utils.mlp import MLP
import copy
import numpy as np
import torch
from torch import optim
from torch.nn import ReLU, Tanh
from gym.spaces import Box, Discrete
from typing import Union


class DQN(ValueBasedAgent):
    """
    An agent that learn an approximated Q-Function using a neural network.
    This Q-Function is used to find the best action to execute in a given observation.
    """

    name = "DQN"

    def __init__(self, state_space: Union[Box, Discrete], action_space: Union[Box, Discrete], **params):
        """
        @param state_space: Environment's state space.
        @param action_space: Environment's action_space.
        @param params: Optional parameters.
        """

        super().__init__(state_space, action_space, **params)

        self.gamma = params.get("gamma", 0.95)
        self.epsilon_min = params.get("epsilon_min", 0.01)
        self.epsilon_max = params.get("epsilon_max", 1.)
        self.epsilon_decay_delay = params.get("epsilon_decay_delay", 20)
        self.epsilon = None
        self.epsilon_decay_period = params.get("epsilon_decay_period", 1000)

        #  NEW, goals will be stored inside the replay buffer. We need a specific one with enough place to do so
        self.learning_rate = params.get("learning_rate", 0.001)
        self.steps_before_target_update = params.get("steps_before_target_update", 1)
        self.steps_since_last_target_update = 0
        self.tau = params.get("tau", 0.001)
        self.nb_gradient_steps = params.get("nb_gradient_steps", 1)

        self.epsilon_step = (self.epsilon_max - self.epsilon_min) / self.epsilon_decay_period
        self.total_steps = 0

        # NEW, The input observation size is multiplied by two because we need to also take the goal as input
        model_layers = params.get("model_layers", [64, ReLU(), 64, ReLU(), 64, ReLU()])
        model_activation = params.get("model_activation", Tanh())
        assert isinstance(model_activation, torch.nn.Module)
        self.model = MLP(self.state_size, *model_layers, self.nb_actions, model_activation,
                         learning_rate=self.learning_rate, optimizer_class=optim.Adam, device=self.device).float()

        self.criterion = torch.nn.SmoothL1Loss()
        self.target_model = copy.deepcopy(self.model).to(self.device)
        self.epsilon = self.epsilon_max

    def get_value(self, observations, actions=None):
        with torch.no_grad():
            values = self.model(observations)
            if actions is None:
                values = values.max(-1).values
            else:
                if not isinstance(actions, torch.Tensor):
                    actions = torch.tensor(actions)
                values = values.gather(1, actions.to(torch.long).unsqueeze(1))
        return values.detach().numpy()

    def action(self, state, explore=True):
        if self.simulation_time_step_id > self.epsilon_decay_delay:
            self.epsilon = max(self.epsilon_min, self.epsilon - self.epsilon_step)

        if explore and not self.under_test and np.random.rand() < self.epsilon:  # Epsilon greedy
            action = np.random.randint(self.nb_actions)
        else:
            # greedy_action(self.model, observation) function in RL5 notebook
            with torch.no_grad():
                q_values = self.model(state)
                action = torch.argmax(q_values).item()

        action = scale_tensor(action, Box(-1, 1, (self.nb_actions,)), self.action_space)
        return action

    def learn(self):
        assert not self.under_test
        for _ in range(self.nb_gradient_steps):
            if len(self.replay_buffer) > self.batch_size:
                states, actions, rewards, new_states, dones = self.sample_training_batch()
                q_prime = self.target_model(new_states).max(1)[0].detach()
                update = rewards + self.gamma * (1 - dones) * q_prime
                q_s_a = self.model(states).gather(1, actions.to(torch.long).unsqueeze(1))
                loss = self.criterion(q_s_a, update.unsqueeze(1))
                self.model.learn(loss)

        self.steps_since_last_target_update += 1
        if self.steps_since_last_target_update >= self.steps_before_target_update:
            self.target_model.converge_to(self.model, self.tau)
            self.steps_since_last_target_update = 0



import pickle
from copy import deepcopy

import numpy as np
import torch
from gym.spaces import Box, Discrete
from typing import Union
from torch.nn import ReLU, Tanh

from hbrl.utils import create_dir
from ..utils import scale_tensor
from ..value_based_agent import ValueBasedAgent
from ..utils.mlp import MLP
from torch import optim


class DDPG(ValueBasedAgent):
    name = "DDPG"

    def __init__(self, state_space: Union[Box, Discrete], action_space: Union[Box, Discrete], **params):
        """
        @param state_space: Environment's state space.
        @param action_space: Environment's action_space.
        @param params: Optional parameters.
        """

        super().__init__(state_space, action_space, **params)

        self.actor_lr = params.get("actor_lr", 0.000025)
        self.critic_lr = params.get("critic_lr", 0.00025)
        self.tau = params.get("tau", 0.001)
        self.gamma = params.get("gamma", 0.99)
        self.noise_std = params.get("noise_std", 0.1)
        self.steps_before_target_update = params.get("steps_before_target_update", 5)
        self.steps_since_last_update = 0

        actor_layers = params.get("actor_layers", [64, ReLU(), 64, ReLU(), 64, ReLU()])
        actor_activation = params.get("actor_activation", Tanh())
        assert isinstance(actor_activation, torch.nn.Module)
        self.actor = MLP(self.state_size, *actor_layers, self.nb_actions, actor_activation,
                         learning_rate=self.actor_lr, optimizer_class=optim.Adam, device=self.device).float()
        critic_layers = params.get("critic_layers", [64, ReLU(), 64, ReLU(), 64, ReLU()])
        self.critic = MLP(self.state_size + self.nb_actions, *critic_layers, 1, learning_rate=self.critic_lr,
                          optimizer_class=optim.Adam, device=self.device).float()

        self.target_actor = deepcopy(self.actor)
        self.target_critic = deepcopy(self.critic)

        self.normal_distribution = torch.distributions.normal.Normal(
            torch.zeros(self.nb_actions), torch.full((self.nb_actions,), self.noise_std))

    def action(self, observation, explore=True):
        with torch.no_grad():
            observation = torch.tensor(observation, dtype=torch.float).to(self.device)
            action = self.actor(observation).to(self.device)
            if not self.under_test and explore:
                action += self.normal_distribution.sample()
            action = action.cpu().detach().numpy()

            # Fit action to our action_space
            action = scale_tensor(action, Box(-1, 1, (self.nb_actions,)), self.action_space)
        return action

    def learn_interaction(self, *interaction_data):
        assert not self.under_test
        self.replay_buffer.append(interaction_data)

    def get_value(self, observations, actions=None):
        with torch.no_grad():
            if actions is None:
                actions = self.actor(observations)
            if isinstance(observations, np.ndarray):
                observations = torch.Tensor(observations)
            if isinstance(actions, np.ndarray):
                actions = torch.Tensor(actions)
            critic_value = self.critic(torch.concat((observations, actions), dim=-1))
        return critic_value.flatten().detach().numpy()

    def learn(self):
        assert not self.under_test
        if not self.under_test and len(self.replay_buffer) > self.batch_size:
            states, actions, rewards, new_states, dones = self.sample_training_batch()

            with torch.no_grad():
                target_actions = self.target_actor(new_states)
                target_actions = scale_tensor(target_actions, Box(-1, 1, (self.nb_actions,)), self.action_space)
                critic_value_ = self.target_critic(torch.concat((new_states, target_actions), dim=-1))
            critic_value = self.critic(torch.concat((states, actions), dim=-1))
            #target = torch.addcmul(rewards, self.gamma, 1 - dones, critic_value_.squeeze()).view(self.batch_size, 1)
            target = (rewards + self.gamma * (1 - dones) * critic_value_.squeeze()).view(self.batch_size, 1)
            critic_loss = torch.nn.functional.mse_loss(target, critic_value)
            self.critic.learn(critic_loss)

            actions = self.actor(states)
            actions = scale_tensor(actions, Box(-1, 1, (self.nb_actions,)), self.action_space)
            actor_loss = - self.critic(torch.concat((states, actions), dim=-1))
            actor_loss = torch.mean(actor_loss)
            self.actor.learn(actor_loss)

            self.steps_since_last_update += 1
            if self.steps_since_last_update % self.steps_before_target_update == 0:
                self.target_critic.converge_to(self.critic, tau=self.tau)
                self.target_actor.converge_to(self.actor, tau=self.tau)

    def save(self, directory):
        if directory[-1] != "/":
            directory += "/"
        create_dir(directory)

        torch.save(self.critic, directory + "critic.pt")
        torch.save(self.target_critic, directory + "target_critic.pt")

        torch.save(self.actor, directory + "actor.pt")
        torch.save(self.target_actor, directory + "target_actor.pt")

        with open(directory + "replay_buffer.pkl", "wb") as f:
            pickle.dump(self.replay_buffer, f)

    def load(self, directory):
        if directory[-1] != "/":
            directory += "/"
        directory = directory[:-1]
        directory += "/"
        create_dir(directory)

        self.critic = torch.load(directory + "critic.pt")
        self.target_critic = torch.load(directory + "target_critic.pt")

        self.actor = torch.load(directory + "actor.pt")
        self.target_actor = torch.load(directory + "target_actor.pt")

        with open(directory + "replay_buffer.pkl", "rb") as f:
            self.replay_buffer = pickle.load(f)



import copy
import json
from typing import Union
import abc
import gym
from gym.spaces import Box, Discrete
import numpy as np
import torch


class Agent(abc.ABC):
    """
    A global agent class that describe the interactions between our agent, and it's environment.
    """

    name = "Agent"

    def __init__(self, state_space: Union[Box, Discrete], action_space: Union[Box, Discrete], **params):
        """
        @param state_space: Environment's state space.
        @param action_space: Environment's action_space.
        @param params: Optional parameters.
        """
        assert isinstance(state_space, Box) or isinstance(state_space, Discrete), \
            "The state space should be an instance of gym.spaces.Space"
        assert isinstance(action_space, Box) or isinstance(action_space, Discrete), \
            "The action space should be an instance of gym.spaces.Space"

        self.init_params = params
        self.state_space = state_space
        self.action_space = action_space
        self.device = params.get("device", torch.device("cpu"))

        # Mandatory parameters
        assert not isinstance(self.state_space, dict), "state space as dictionary is not supported."
        self.state_size = self.state_space.shape[0]  # Assume observation space is continuous
        self.state_shape = self.state_space.shape
        assert len(self.state_shape) == 1

        self.continuous = isinstance(self.action_space, gym.spaces.Box)
        self.nb_actions = self.action_space.shape[0] if self.continuous else self.action_space.n
        self.last_state = None  # Useful to store interaction when we receive (new_stare, reward, done) tuple
        self.episode_id = 0
        self.episode_time_step_id = 0
        self.simulation_time_step_id = 0
        self.output_dir = None
        self.sub_plots_shape = ()
        self.under_test = False
        self.episode_started = False

    def start_episode(self, state: np.ndarray, test_episode=False):
        if self.episode_started:
            self.stop_episode()
        self.episode_started = True
        self.last_state = state
        self.episode_time_step_id = 0
        self.under_test = test_episode

    @abc.abstractmethod
    def action(self, state, explore=True):
        pass

    def process_interaction(self, action, reward, new_state, done, learn=True):
        self.episode_time_step_id += 1
        self.simulation_time_step_id += 1
        self.last_state = new_state

    def stop_episode(self):
        self.episode_id += 1
        self.episode_started = False

    def reset(self):
        self.__init__(self.state_space, self.action_space, **self.init_params)

    def copy(self):
        return copy.deepcopy(self)

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if isinstance(v, Agent):
                setattr(result, k, v.copy())
            elif isinstance(v, dict):
                new_dict = {}
                for k_, v_ in v.items():
                    new_dict[k_] = v_.copy() if k_ == "goal_reaching_agent" else copy.deepcopy(v_)
                setattr(result, k, new_dict)
            else:
                setattr(result, k, copy.deepcopy(v, memo))
        return result

    def save(self, load_dir):
        pass

    def load(self, load_dir):
        pass


from enum import Enum

ant_maze = False
try:
    import mujoco_py
    ant_maze = True
except ImportError:
    pass

class EnvironmentIndex(Enum):
    GRID_WORLD = "grid_world"
    POINT_MAZE = "point_maze"
    if ant_maze:
        ANT_MAZE = "ant_maze"
